# pylint: skip-file
from .base import CanonicalException
from .schema import CollectionValidationErrorSchema
from .schema import ValidationErrorSchema

try:
    from . import pydantic
except ImportError:
    pass


class DoesNotExist(CanonicalException):
    """Raised to indicate that a resource lookup was requested but no
    entity matched the given predicate.
    """
    code = 'RESOURCE_DOES_NOT_EXIST'
    message = "The requested resource does not exist."
    detail = "No resources were found that matched the given search query."
    http_status_code = 404


class Duplicate(CanonicalException):
    """Raised to indicate that the creation of a resource was requested,
    but there was a conflicting identifier.
    """
    code = 'RESOURCE_EXISTS'
    message = "A resource with the provided identifying attributes exists."
    http_status_code = 409


class MultipleObjectsReturned(CanonicalException):
    """Raised when a single resource was requested but the search predicate
    yielded multiple results.
    """
    http_status_code = 404
    code = 'RESOURCE_DOES_NOT_EXIST'
    message = "The requested resource does not exist."
    detail = "Multiple resources were found that matched the given search query."
    hint = "Limit the search query, if possible."


class ValidationError(CanonicalException):
    """Raised when input data did not satisfy the specifed format
    requirements.
    """
    schema_class = ValidationErrorSchema
    http_status_code = 422
    code = 'DATA_CONSTRAINT_VIOLATION'
    message = "The input data did not satisfy the specification."
    hint = "Inspect .spec to see field-specific errors."

    def __init__(self, *args, **kwargs):
        self.spec = kwargs.pop('spec')
        self.index = kwargs.pop('index', None)
        super().__init__(*args, **kwargs)



class CollectionValidationError(CanonicalException):
    """Raised when collection input data did not satisfy the specifed format
    requirements.
    """
    schema_class = CollectionValidationErrorSchema
    http_status_code = 422
    code = 'DATA_CONSTRAINT_VIOLATION'
    message = "The input data did not satisfy the specification."
    hint = "Inspect .items to see the errors for each item."

    def __init__(self, *args, **kwargs):
        self.items = kwargs.pop('items')
        super().__init__(*args, **kwargs)


class FeatureNotSupported(CanonicalException):
    """Raised when a feature is requested that is not supported e.g. because
    of misconfiguration, deprecation or retirement.
    """
    http_status_code = 503
    code = 'FEATURE_NOT_SUPPORTED'
    message = "The requested feature is not supported."
    detail = (
        "This feature is either not configured, deprecated or retired. Consult "
        "the API documentation for more information."
    )


class BaseTokenException(CanonicalException):

    def __init__(self, *args, **kwargs):
        self.token = kwargs.pop('token')
        super().__init__(*args, **kwargs)


class InvalidToken(CanonicalException):
    code = 'BEARER_TOKEN_INVALID'
    http_status_code = 403
    message = "The authentication token is not valid."


class InvalidScope(BaseTokenException):
    code = 'BEARER_TOKEN_NOT_AUTHORIZED'
    http_status_code = 403
    message = "The provided token has an invalid scope."


class InvalidAudience(BaseTokenException):
    code = 'BEARER_TOKEN_AUDIENCE_INVALID'
    http_status_code = 403
    message = "The audience of the authentication token is invalid."


class ExpiredSignatureError(BaseTokenException):
    code = 'BEARER_TOKEN_SIGNATURE_EXPIRED'
    http_status_code = 403
    message = "The authentication token is expired."


class InvalidSignatureError(BaseTokenException):
    code = 'BEARER_TOKEN_SIGNATURE_INVALID'
    http_status_code = 403
    message = "The signature of the bearer token could not be verified."


class MissingToken(BaseTokenException):
    code = 'BEARER_TOKEN_MISSING'
    http_status_code = 403
    message = "The authentication token is missing."
    hint = "Provide the Bearer token using the Authorization header."


class CrossSiteRequestForgeryTokenExpired(BaseTokenException):
    code = 'CSRF_TOKEN_EXPIRED'
    http_status_code = 403
    message = "The presented CSRF token is expired."
    hint = "Request a new token."


class CrossSiteRequestForgeryTokenMissing(BaseTokenException):
    code = 'CSRF_TOKEN_MISSING'
    http_status_code = 403
    message = "The presented CSRF token is missing."
    hint = "Request a new token."


class CrossSiteRequestForgeryTokenInvalid(BaseTokenException):
    code = 'CSRF_TOKEN_INVALID'
    http_status_code = 403
    message = "The presented CSRF token is not valid."
    hint = "Request a new token."



"""Declares :class:`Repository`."""
import abc

from unimatrix.lib.datastructures import ImmutableDTO
from .exc import Duplicate
from .exc import DoesNotExist
from .exc import MultipleObjectsReturned


class Repository(metaclass=abc.ABCMeta):
    """The base class for all :term:`Repository` implementations."""
    dto_class = ImmutableDTO

    #: Raised when a new entity with an existing key is being persisted.
    Duplicate = Duplicate

    #: Raised when an entity is requested with a search predicate that
    #: yielded no result.
    DoesNotExist = DoesNotExist

    #: Raised when an entity is requested but the search predicated yielded
    #: more than one result.
    MultipleObjectsReturned = MultipleObjectsReturned

    @classmethod
    def class_factory(cls, impl):
        """Create a new :class:`Repository` subclass using the
        provided implementation class `impl`.
        """
        return type('%sRepository' % impl.__name__, (impl, cls), {
            '__module__': impl.__module__,
            '_in_transaction': False
        })

    @classmethod
    def new(cls, impl, *args, **kwargs):
        """Create a new instance with the specified implementation class."""
        return cls.class_factory(impl)(*args, **kwargs)

    def __init__(self):
        self._in_transaction = False

    async def list(self, limit=100, offset=0, **kwargs):
        """Return a list of references to domain object based on the provided
        search predicate(s).
        """
        raise NotImplementedError
    list.not_implemented = True

    async def exists(self, *args, **kwargs):
        """Return a boolean indicating if any domain object with properties
        that match the search predicate exists.
        """
        raise NotImplementedError
    exists.not_implemented = True

    async def get(self, key, *args, **kwargs):
        """Lookup a domain object using the given identifying `key`."""
        raise NotImplementedError
    get.not_implemented = True

    async def get_as_dto(self, key, *args, **kwargs):
        """Like :meth:`get`, but returns a Data Transfer Object (DTO). Use
        this method when transferring complete domain objects to external
        systems or interfaces.
        """
        raise NotImplementedError
    get_as_dto.not_implemented = True

    async def persist(self, instance):
        """Persist a domain object to the underlying storage backend."""
        raise NotImplementedError
    persist.not_implemented = True

    async def persist_from_dto(self, dto):
        """Like :meth:`persist`, but takes a Data Transfer Object (DTO) as
        its input. Use this method when receiving a serialized domain object
        from an external system or interface.
        """
        raise NotImplementedError
    persist_from_dto.not_implemented = True


    def as_context(self, *args, **kwargs):
        """Hook that is executed when a transaction is started."""
        if self._in_transaction:
            raise RuntimeError("Nested transactions are not supported.")
        repo = type(self)()
        repo._in_transaction = True
        repo.setup_context(*args, **kwargs)
        return repo

    def with_context(self, *args, **kwargs):
        """Hook that is executed when a transaction is started."""
        return self.as_context(*args, **kwargs)

    def setup_context(self, *args, **kwargs):
        """Hook to setup the context."""
        pass

    def teardown_context(self, cls, exc, tb):
        """Hook to teardown the context."""
        pass

    def __enter__(self):
        return self

    def __exit__(self, cls, exc, tb):
        pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, cls, exc, tb):
        self.async_teardown_context(cls, exc, tb)



# pylint: skip-file
import unittest

from ..base import CanonicalException


class LoggingTestCase(unittest.TestCase):

    def test_log_message_formatting(self):
        exc = CanonicalException()
        msg = exc.log(lambda x: x)
        self.assertEqual("Caught fatal %s (id: %s)" % (exc.code, str(exc.id)), msg) 

    def test_log_message_formatting_custom(self):
        exc = CanonicalException()
        msg = exc.log(lambda x: x, 'foo')
        self.assertEqual("foo", msg) 



"""Declares exception types."""
import uuid

from .schema import CanonicalExceptionSchema


class CanonicalException(Exception):
    """The base class for all exceptions."""

    #: The schema class used to convert the canonical exception to a
    #: dictionary.
    schema_class = CanonicalExceptionSchema

    #: The HTTP default HTTP status code.
    http_status_code = None

    #: The default code for the exception.
    code = 'UNCAUGHT_EXCEPTION'

    #: The default message.
    message = None

    #: The default detail.
    detail = None

    #: The default hint
    hint = None

    #: The default log message. String formatting is applied.
    log_message = "Caught fatal {code} (id: {id})"

    def __init__(self, code=None, message=None, detail=None, hint=None,
        http_status_code=None):
        """Initialize a new :class:`CanonicalException`."""
        self.id = uuid.uuid4()
        self.code = code or self.code
        self.message = message or self.message
        self.detail = detail or self.detail
        self.hint = hint or self.hint
        self.http_status_code = http_status_code or self.http_status_code
        self.schema = self.schema_class()

    def as_dict(self):
        """Return a dictionary containing the exception properties."""
        return self.schema.dump(self)

    def log(self, func, message=None):
        """Use `logger` to log the exception as `level`."""
        return func((message or self.log_message).format(**self.as_dict()))



# pylint: skip-file
import unittest

from .. import ValidationError
from ..schema import CanonicalExceptionSchema


class ValidationErrorTestCase(unittest.TestCase):

    def test_create_from_marshmallow_exception(self):
        schema = CanonicalExceptionSchema()
        try:
            schema.load({})
        except Exception as e:
            exc = ValidationError(spec=e.normalized_messages())
            dto = exc.as_dict()
            self.assertIn('spec', dto)
        else:
            self.fail()



"""Declares :mod:`pydantic` models for all exception classes."""
from typing import List
from typing import Optional

from pydantic import BaseModel


class CanonicalException(BaseModel):
    """Declares the :mod:`pydantic` schema for canonical exceptions."""

    code: str
    message: str
    detail: Optional[str]
    key: Optional[List]



# pylint: skip-file
import unittest

from .. import Repository


class ClassCreationTestCase(unittest.TestCase):

    def test_create_class(self):

        class Impl:
            pass

        repo = Repository.new(Impl)

    def test_with_context(self):

        class Impl:

            def setup_context(self, foo):
                self.foo = foo

        repo = Repository.new(Impl)
        with repo.with_context(foo=1) as tmp:
            self.assertEqual(tmp.foo, 1)

    def test_with_context_can_not_be_nested(self):

        class Impl:

            def setup_context(self, foo):
                self.foo = foo

        repo = Repository.new(Impl)
        with repo.with_context(foo=1) as tmp:
            with self.assertRaises(RuntimeError):
                with tmp.with_context(foo=1) as tmp:
                    self.fail()



"""Declares :class:`ValidationErrorSchema`."""
import marshmallow
import marshmallow.fields

from .base import BaseSchema


class ValidationErrorSchema(BaseSchema, marshmallow.Schema):
    """The schema for :class:`unimatrix.ext.model.exc.ValidationError`
    exceptions.
    """

    spec = marshmallow.fields.Dict(
        required=True,
        keys=marshmallow.fields.String(),
        values=marshmallow.fields.List(marshmallow.fields.String)
    )

    index = marshmallow.fields.Integer()

    @marshmallow.post_dump
    def remove_index(self, data, *args, **kwargs):
        if data.get('index') is None:
            data.pop('index')
        return data


class CollectionValidationErrorSchema(BaseSchema, marshmallow.Schema):
    """The schema for :class:`unimatrix.ext.model.exc.CollectionValidationError`
    exceptions.
    """

    items = marshmallow.fields.List(
        marshmallow.fields.Nested(ValidationErrorSchema),
        required=True
    )



# pylint: skip-file
import unittest

from unimatrix.ext.model import exc


class ExceptionSchemaTestCase(unittest.TestCase):

    def test_serialize_canonical_exception_has_id(self):
        e = exc.CanonicalException()
        dto = e.as_dict()
        self.assertEqual(str(e.id), dto['id'])

    def test_serialize_canonical_exception_has_code(self):
        e = exc.CanonicalException()
        dto = e.as_dict()
        self.assertEqual(e.code, dto['code'])

    def test_serialize_canonical_exception_has_message(self):
        e = exc.CanonicalException()
        dto = e.as_dict()
        self.assertEqual(e.message, dto['message'])

    def test_serialize_canonical_exception_has_detail(self):
        e = exc.CanonicalException()
        dto = e.as_dict()
        self.assertEqual(e.detail, dto['detail'])

    def test_serialize_canonical_exception_has_hint(self):
        e = exc.CanonicalException()
        dto = e.as_dict()
        self.assertEqual(e.hint, dto['hint'])



# pylint: skip-file
import unittest

from .. import ValidationError
from .. import CollectionValidationError
from ..schema import CanonicalExceptionSchema


class CollectionValidationErrorTestCase(unittest.TestCase):

    def test_create_from_marshmallow_exception(self):
        schema = CanonicalExceptionSchema()
        try:
            schema.load({})
        except Exception as e:
            items = [
                ValidationError(spec=e.normalized_messages(), index=2),
                ValidationError(spec=e.normalized_messages(), index=4)
            ]
            exc = CollectionValidationError(items=items)
            dto = exc.as_dict()
            self.assertIn('items', dto)
            self.assertEqual(dto['items'][0]['index'], 2)
            self.assertEqual(dto['items'][1]['index'], 4)
        else:
            self.fail()



"""Declares :class:`CanonicalExceptionSchema`."""
import marshmallow
import marshmallow.fields

from .base import BaseSchema


class CanonicalExceptionSchema(BaseSchema, marshmallow.Schema):
    """The schema for exceptions that are not otherwise specified."""



"""Declares :class:`BaseSchema`."""
import marshmallow.fields


class BaseSchema:
    id = marshmallow.fields.UUID(
        required=True
    )

    code = marshmallow.fields.String(
        required=True
    )

    message = marshmallow.fields.String(
        required=True
    )

    detail = marshmallow.fields.String(
        required=False
    )

    hint = marshmallow.fields.String(
        required=False
    )
